# Config command data frame
CMD_HEADER = "FDFCFBFA"
CMD_MFR = "04030201"

# Command Constants
CMD_CONFIG_ENABLE = "0400FF000100"
CMD_CONFIG_DISABLE = "0200FE00"
CMD_PARAM_EDIT = "14006000"
CMD_PARAM_READ = "02006100"
CMD_ENG_MODE_ENABLE = "02006200"
CMD_ENG_MODE_DISABLE = "02006300"
CMD_GATE_SENS_EDIT = "14006400"
CMD_FIRMWARE_READ = "0200A000"
CMD_BAUD_RATE_SET = "0400A100"
CMD_FACTORY_RESET = "0200A200"
CMD_RESTART = "0200A300"
CMD_BT_ENABLE = "0400A4000100"
CMD_BT_DISABLE = "0400A4000000"
CMD_BT_MAC_QUERY = "0400A5000100"

# Read constants
REF_READ_HEADER = "F4F3F2F1"

# Parameter Constants

# Configure base settings
PARAM_MAX_MOVING_GATE = "0000"
PARAM_MAX_STATIC_GATE = "0100"
PARAM_EMPTY_DURATION = "0200"
# Configure individual gates
PARAM_GATE_SELECT = "0000"
PARAM_MOVING_GATE_WORD = "0100"
PARAM_STATIC_GATE_WORD = "0200"
# Special case
PARAM_SELECT_ALL_GATE = "FFFF"

# Set Baud Rate
PARAM_BAUD_9600 = "0100"
PARAM_BAUD_19200 = "0200"
PARAM_BAUD_38400 = "0300"
PARAM_BAUD_57600 = "0400"
PARAM_BAUD_115200 = "0500"
PARAM_BAUD_230400 = "0600"
PARAM_BAUD_256000 = "0700"
PARAM_BAUD_460800 = "0800"

PARAM_ACCEPTABLE_BAUDS = [PARAM_BAUD_9600  ,
                          PARAM_BAUD_19200 ,
                          PARAM_BAUD_38400 ,
                          PARAM_BAUD_57600 ,
                          PARAM_BAUD_115200,
                          PARAM_BAUD_230400,
                          PARAM_BAUD_256000,
                          PARAM_BAUD_460800]

PARAM_DEFAULT_BAUD = PARAM_BAUD_256000

# I know, it's very hacky lol
BAUD_LOOKUP = {PARAM_BAUD_9600  : 9600,
             PARAM_BAUD_19200 : 19200,
             PARAM_BAUD_38400 : 38400,
             PARAM_BAUD_57600 : 57600,
             PARAM_BAUD_115200: 115200,
             PARAM_BAUD_230400: 230400,
             PARAM_BAUD_256000: 256000,
             PARAM_BAUD_460800: 460800}

# Ack Constants
ACK_CONFIG_ENABLE = "0800FF01000001004000"

# Validation Constants
MAX_BUFFER_SIZE = 64 # 32 byte buffer read
GATE_MIN = 0
GATE_MAX = 8
TIMEOUT_MIN = 0
TIMEOUT_MAX = 65535
SENS_MIN = 0
SENS_MAX = 100

# Read Parameter Indices
REF_MAX_MOVING_GATE = 12
REF_MAX_STATIC_GATE = 13

REF_MOVING_GATE_SENS_0 = 14
REF_MOVING_GATE_SENS_1 = 15
REF_MOVING_GATE_SENS_2 = 16
REF_MOVING_GATE_SENS_3 = 17
REF_MOVING_GATE_SENS_4 = 18
REF_MOVING_GATE_SENS_5 = 19
REF_MOVING_GATE_SENS_6 = 20
REF_MOVING_GATE_SENS_7 = 21
REF_MOVING_GATE_SENS_8 = 22

REF_STATIC_GATE_SENS_0 = 23
REF_STATIC_GATE_SENS_1 = 24
REF_STATIC_GATE_SENS_2 = 25
REF_STATIC_GATE_SENS_3 = 26
REF_STATIC_GATE_SENS_4 = 27
REF_STATIC_GATE_SENS_5 = 28
REF_STATIC_GATE_SENS_6 = 29
REF_STATIC_GATE_SENS_7 = 30
REF_STATIC_GATE_SENS_8 = 31

REF_EMPTY_TIMEOUT = 32

REF_FW_MAJOR_HEAD = 12
REF_FW_MAJOR_TAIL = 14
REF_FW_MINOR_HEAD = 14
REF_FW_MINOR_TAIL = 18

REF_BT_ADDR_HEAD = 10
REF_BT_ADDR_TAIL = 16

REF_NORMAL_PACKET_LEN = 15
REF_ENG_MODE_PACKET_LEN = 37

REF_PACKET_CRC = "5500"

REF_TARGET_TYPE = 4

REF_TARGET_LOOKUP = {0: "No Target",
                     1: "Moving Target",
                     2: "Static Target",
                     3: "Moving + Static Target"}

REF_TARGET_MOVE_DIST_HEAD = 5
REF_TARGET_MOVE_DIST_TAIL = 6
REF_TARGET_MOVE_ENERGY = 7
REF_TARGET_STATIC_DIST_HEAD = 8
REF_TARGET_STATIC_DIST_TAIL = 9
REF_TARGET_STATIC_ENERGY = 10
REF_DETECT_DIST = 11

# Cut out 1-7 for brevity
REF_MOVING_GATE_ENERGY_0 = 14
REF_MOVING_GATE_ENERGY_8 = 23

REF_STATIC_GATE_ENERGY_0 = 24
REF_STATIC_GATE_ENERGY_8 = 33

REF_RADAR_POLING_RATE = 10 # Radar updates at 10Hz

REF_PACKET_CRC_IDX = -2

REF_ENG_CHECK_IDX = 0
REF_ENG_CHECK = 35


from .ld2410_consts import *
from collections import deque
import serial
import struct
import threading
import time
import logging
import signal


class Queue:
    def __init__(self, max_size):
        self.queue = deque(maxlen=max_size)
    
    def add(self, element):
        self.queue.append(element)
    
    def byte_str(self):
        return b"".join(list(self.queue))



class LD2410():
    def __init__(self, port, baud_rate=PARAM_DEFAULT_BAUD, timeout=1, verbosity=logging.DEBUG) -> None:
        self.port = port
        self.baudrate = baud_rate
        self.timeout = timeout
        self.eng_mode = False
        self.read_fail_count = 0
        
        # threading data variables
        self.last_detection = None
        self._lock = threading.Lock()
        self._worker_thread = None
        self._stop_event = threading.Event()

        logging.basicConfig(level=verbosity)

        self.ser = serial.Serial(port, BAUD_LOOKUP[baud_rate], timeout=timeout)
        logging.info(f"Serial port initialised at {port}, with baud rate {BAUD_LOOKUP[baud_rate]}")

    # Validate that the input is within a valid range
    @staticmethod
    def validate_range(input, lower, upper):
        if input in range(lower, upper):
            return True
        else:
            input_name = f'{input=}'.split('=')[0]
            raise Exception(f"{input_name} {input} is not a valid setting, please pick a value between {lower} and {upper}")

    # Helper functions
    @staticmethod
    def frame_wrapper(command):
        return bytes.fromhex(CMD_HEADER + command + CMD_MFR)

    # Convert a decimal integer to a 4 byte little endian string
    @staticmethod
    def int_to_4b(num):
        hex_string = bytearray.fromhex(struct.pack('>I', num).hex())
        hex_string.reverse()
        return bytes(hex_string).hex()

    # Base functions

    # Sends a dataframe encoded as bytes enclosed within a format specific header
    # Returns the acknowledged data received from the radar
    def send_frame(self, command):
        while True:
            try:
                # Wrap up the command 
                command_bytes = self.frame_wrapper(command)
                
                self.ser.write(command_bytes)
                ret_bytes = self.ser.read(MAX_BUFFER_SIZE)
                
                logging.debug(f"Sending data:  {command_bytes.hex(' ')}")
                logging.debug(f"Received data: {ret_bytes.hex(' ')}")
    
                return ret_bytes # Returns the response given by the radar module
            
            except Exception as e:
                logging.debug(e)
    
    def send_command(self, command):
        # Enable config mode
        self.send_frame(CMD_CONFIG_ENABLE)
        # Send command
        ret_bytes = self.send_frame(command)
        # Disable config mode
        self.send_frame(CMD_CONFIG_DISABLE)

        return ret_bytes
        

    # Configure Detection Gates and Detect Duration
    def edit_detection_params(self, moving_max_gate, static_max_gate, timeout):
        logging.info("Editing detection parameters")
        self.validate_range(moving_max_gate, GATE_MIN, GATE_MAX+1)
        self.validate_range(static_max_gate, GATE_MIN, GATE_MAX+1)
        self.validate_range(timeout, TIMEOUT_MIN, TIMEOUT_MAX+1)

        command = CMD_PARAM_EDIT + PARAM_MAX_MOVING_GATE + self.int_to_4b(moving_max_gate) \
                      + PARAM_MAX_STATIC_GATE + self.int_to_4b(static_max_gate) \
                      + PARAM_EMPTY_DURATION + self.int_to_4b(timeout)
        
        self.send_command(command)

    # Read the currently configured parameters 
    #
    # Returns 3 arrays, 
    # 
    # first array: [moving gate threshold, static gate threshold, empty timeout]
    # second array: [moving gate sens 0.... moving gate sens 8]
    # third array: [static gate sens 0.... static gate sens 8]

    def read_detection_params(self):
        # Send command to retrieve parameters
        logging.info("Reading detection parameters")
        ret = self.send_command(CMD_PARAM_READ)
        
        # Process response
        # Threshold Params
        thresholds = [ret[REF_MAX_MOVING_GATE], ret[REF_MAX_STATIC_GATE], ret[REF_EMPTY_TIMEOUT]]


        # Movement Gate Sensitivities
        move_sens = [int(byte) for byte in ret[REF_MOVING_GATE_SENS_0:REF_MOVING_GATE_SENS_8]]
        # Static Gate Sensitivities
        static_sens = [int(byte) for byte in ret[REF_STATIC_GATE_SENS_0:REF_STATIC_GATE_SENS_8]]
        
        logging.debug(f"Thresholds:{thresholds}, Movement Sens:{move_sens}, Static Sens:{static_sens}")
        return thresholds, move_sens, static_sens
        
    # Enable Engineering Mode
    # Adds energy level of each gate to the radar output
    def enable_engineering_mode(self):
        logging.info("Enabling engineering mode")
        self.eng_mode = True
        self.send_command(CMD_ENG_MODE_ENABLE)
        
    # Disable Engineering Mode
    def disable_engineering_mode(self):
        logging.info("Disabling engineering mode")
        self.eng_mode = False
        self.send_command(CMD_ENG_MODE_DISABLE)

    # Configure Gate Movement and Static Sensitivities
    def edit_gate_sensitivity(self, gate, moving_sens, static_sens):
        logging.info("Editing gate sensitivity")
        self.validate_range(gate, GATE_MIN, GATE_MAX+1)
        self.validate_range(moving_sens, SENS_MIN, SENS_MAX+1)

        if gate == 1 or gate == 2:
            logging.warning("You cannot set gate 1 or 2 static sensitivity to anything other than 0")
            self.validate_range(static_sens, 0, 1)
        else:
            self.validate_range(static_sens, SENS_MIN, SENS_MAX+1)

        command = CMD_GATE_SENS_EDIT + PARAM_GATE_SELECT + self.int_to_4b(gate) \
                      + PARAM_MOVING_GATE_WORD + self.int_to_4b(moving_sens) \
                      + PARAM_STATIC_GATE_WORD + self.int_to_4b(static_sens)
        
        self.send_command(command)

    # Read Firmware Version
    def read_firmware_version(self):
        logging.info("Reading firmware version")
        ret = self.send_command(CMD_FIRMWARE_READ)

        # Need to flip from little endian to big endian
        fw_major = bytes(reversed(ret[REF_FW_MAJOR_HEAD:REF_FW_MAJOR_TAIL]))
        fw_minor = bytes(reversed(ret[REF_FW_MINOR_HEAD:REF_FW_MINOR_TAIL]))

        fw_version = f"V{fw_major[0]}.{fw_major[1]:02}.{fw_minor.hex()}"
        
        return fw_version
    
    # Used to set the baud rate of the module. 
    # Setting "reconnect=True" will allow the driver to 
    # restart and reset the module and adapt the host to the new baud rate

    # Check consts.py to find the right settings
    def set_baud_rate(self, baud_rate, reconnect=True):
        logging.info(f"Setting baud rate to {BAUD_LOOKUP[baud_rate]}")
        if baud_rate not in PARAM_ACCEPTABLE_BAUDS:
            raise Exception(f"{baud_rate} is not a valid setting. Consult consts.py to find an appropriate setting.")
        
        self.send_command(CMD_BAUD_RATE_SET + baud_rate)
        
        if reconnect:
            logging.info("Baud rate set command issued. Calling restart.")
            # Restart the driver with the new baudrate
            self.restart_module(BAUD_LOOKUP[baud_rate])

    def factory_reset(self, reconnect=True):
        logging.warning("Module will now be factory reset")
        self.send_command(CMD_FACTORY_RESET)
        if reconnect:
            self.restart_module(BAUD_LOOKUP[PARAM_DEFAULT_BAUD])

    def restart_module(self, new_baud=None):
        logging.info("Restarting module")
        if new_baud:
            self.baudrate = new_baud

        self.send_command(CMD_RESTART)
        self.ser.close()
        self.ser = self.ser = serial.Serial(self.port, BAUD_LOOKUP[self.baudrate], timeout=self.timeout)
        self.eng_mode = False

        time.sleep(1)

    # Enable Bluetooth
    def bt_enable(self):
        logging.info("Enabling Bluetooth")
        self.send_command(CMD_BT_ENABLE)

    # Disable Bluetooth
    def bt_disable(self):
        logging.info("Disabling Bluetooth")
        self.send_command(CMD_BT_DISABLE)

    # Get Bluetooth MAC Address
    # Returns a string in the format of xx:xx:xx:xx:xx:xx
    def bt_query_mac(self):
        logging.info("Getting Bluetooth Address")
        ret = self.send_command(CMD_BT_MAC_QUERY)
        mac = ret[REF_BT_ADDR_HEAD:REF_BT_ADDR_TAIL].hex(":") 
        logging.debug(f"Bluetooth address is {mac}")
        return mac
    
    # Get Radar Frame
    def get_data_frame(self):
        logging.debug("Getting raw dataframe")
        # Keep cycling the buffer until frame starts
        buffer = Queue(max_size=4)
        while buffer.byte_str() != bytes.fromhex(REF_READ_HEADER):
            try:
                b = self.ser.read()
            except:
                self.read_fail_count += 1
                logging.debug("Serial failed to read data. Trying again")
                self.read_fail_count += 1
                if self.read_fail_count > 32:
                    logging.warning("Serial failed to read data many times in a row. Please check if the baud rate is correct. Hint: Check the firmware version, if it looks weird, it's probably wrong")
                b = b""
            buffer.add(b)
        
        # Different packet lengths depending on whether engineering mode is on
        if self.eng_mode:
            read_len = REF_ENG_MODE_PACKET_LEN
        else:
            read_len = REF_NORMAL_PACKET_LEN

        try:
            ret_candidate = self.ser.read(read_len)
        except:
            
            logging.debug("Serial failed to read data. Skipping this read")
            return None

        logging.debug(f"get_data_frame() returning {ret_candidate.hex(' ')}")

        # Catch engineering mode not set error
        if ret_candidate[REF_ENG_CHECK_IDX] == REF_ENG_CHECK and self.eng_mode == False: # Engineering mode is on, but not set in driver
            logging.warning("Data seems to be in engineering mode format. However, driver isn't set to use parse engineering mode. Setting it now")
            self.eng_mode = True
        elif ret_candidate[REF_PACKET_CRC_IDX:] != bytes.fromhex(REF_PACKET_CRC):
            logging.warning(f'Checksum not correct received this packet {ret_candidate.hex(" ")}')
            # raise Exception("Checksum of received data is wrong. Data may be corrupted")

        if ret_candidate:
            self.read_fail_count = 0
        
        return ret_candidate

    # Sets 3 lists (standard, move_energies, static_energies). If engineering mode is disabled, second and third list is empty
    def get_radar_data(self):
        logging.debug("Getting raw dataframe")

        # Keep trying until a successful read
        ret = self.get_data_frame()
        while not ret:
            ret = self.get_data_frame()

        move_energies = None
        static_energies = None
        
        target_type = ret[REF_TARGET_TYPE] # 0 No Target, 1 Moving, 2 Static, 3 Static + Moving
        
        moving_target_dist = int(bytes(reversed(ret[REF_TARGET_MOVE_DIST_HEAD:REF_TARGET_MOVE_DIST_TAIL+1])).hex(), 16)
        moving_target_energy = ret[REF_TARGET_MOVE_ENERGY]
        
        static_target_dist = int(bytes(reversed(ret[REF_TARGET_STATIC_DIST_HEAD:REF_TARGET_STATIC_DIST_TAIL+1])).hex(), 16)
        static_target_energy = ret[REF_TARGET_STATIC_ENERGY]

        detection_dist = ret[REF_DETECT_DIST]

        standard_frame = [target_type, moving_target_dist, moving_target_energy, static_target_dist, static_target_energy, detection_dist]
        
        logging.debug(f"Returning dataframes {standard_frame}, {move_energies}, {static_energies}")

        if self.eng_mode:
            # Movement Gate Sensitivities
            move_energies = [int(byte) for byte in ret[REF_MOVING_GATE_ENERGY_0:REF_MOVING_GATE_ENERGY_8+1]]
            # Static Gate Sensitivities
            static_energies = [int(byte) for byte in ret[REF_STATIC_GATE_ENERGY_0:REF_STATIC_GATE_ENERGY_8+1]]

        return standard_frame, move_energies, static_energies

    def poll_radar(self):
        while not self._stop_event.is_set():
            with self._lock:
                self.last_detection = self.get_radar_data()
            time.sleep(0.1)

    def start(self):
        logging.info("Radar polling started")
        self._worker_thread = threading.Thread(target=self.poll_radar)
        self._worker_thread.start()
        self._stop_event.clear()
        time.sleep(1) # Allow for a 1s init time

    def stop(self):
        if not self._stop_event.is_set():
            logging.info("Radar polling stopped")
            self._stop_event.set()
            self._worker_thread.join()
        else:
            logging.debug("Calling stop() but radar isn't running. This is normal.")


    def get_data(self):
        logging.debug(f"get_data returning {self.last_detection}")
        with self._lock:
            if not self.last_detection:
                logging.warning("Data is empty, have you started the radar yet?")
            return self.last_detection


from LD2410 import *
import logging
import time


def main():
    radar = LD2410("/dev/ttyUSB0", PARAM_BAUD_460800, verbosity=logging.DEBUG) # Set desired level of verbosity [DEBUG, INFO, WARNING]
    
    # Get Radar Firmware version
    fw_ver = radar.read_firmware_version()
    print(fw_ver)

    # Set max detection gate for moving to 2, static to 3, and empty timeout to 1s
    radar.edit_detection_params(2, 3, 1)

    # Set the gate 3 moving energy sentivity to 50 and static sensitivity to 40
    # Note: Static sensitivity cannot be set for gate 1 and 2, it must be set to zero e.g (1, 50, 0)
    radar.edit_gate_sensitivity(3, 50, 40)

    # Retrieve the set detection parameters
    # Returns 3 arrays 
    # 1. Gate parameters. we set it earlier to (2, 3, 1)
    # 2. Motion energy sensitivities for gates 0-8 
    # 3. Static energy sensitivities for gates 0-8 
    detection_params = radar.read_detection_params()
    print(detection_params)

    # Get data in standard mode

    # Start the radar polling
    radar.start() # The radar polls asynchronously at 10Hz
    
    # Get 3 data frames 1s apart. If you use a delay less than 0.1, you will get repeat data
    for _ in range(3):
        print(radar.get_data()) # The right 2 arrays will be blank since we are polling in standard mode
        time.sleep(1)
    # Get data in engineering mode

    radar.enable_engineering_mode()
    
    # Get 3 data frames 1s apart. If you use a delay less than 0.1, you will get repeat data
    for _ in range(3):
        print(radar.get_data()) # The right 2 arrays will be blank since we are polling in standard mode
        time.sleep(1)

    radar.disable_engineering_mode()

    radar.stop() # Stop polling the radar

    # Restart the radar
    radar.restart_module()

    # Enable bluetooth (On by default, you don't have to call it)
    radar.bt_enable()

    # Get bluetooth MAC address
    print(radar.bt_query_mac())

    
    # Set module baud rate
    # radar.set_baud_rate(PARAM_BAUD_460800) 

    # Factory reset the module
    # radar.factory_reset()

if __name__ == "__main__":
    main()